"""List of installations of one particular package"""
from collections import UserList
from typing import Optional, List, Set

from ..package_metadata import PackageMetadata


class VersionsMatch(UserList):
    """List of installations of one particular package"""
    display_name: str
    """The name to display to the user."""
    expected_num: Optional[int]
    """The number of packages expected to be found. This prevents false positives where versions appear to match \
    due to some versions not being fetched."""

    def __init__(
            self,
            data: List[PackageMetadata],
            display_name: Optional[str] = None,
            expected_num: Optional[int] = None,
    ):
        """Initialize the dependency match"""
        super().__init__(data)
        self.display_name = display_name
        self.expected_num = expected_num

    @property
    def is_consistent(self) -> bool:
        """Do all versions match?"""
        if len(self.deduplicated_versions) == 1:
            return True
        return False

    @property
    def has_all_expected_items(self) -> bool:
        """Does the number of dependencies match the expected value?"""
        return self.expected_num is None or len(self.data) == self.expected_num

    @property
    def versions(self) -> List[str]:
        """All dependency versions with duplicates removed"""
        return [dependency.current for dependency in self.data if dependency.current is not None]

    @property
    def deduplicated_versions(self) -> Set[str]:
        """All dependency versions with duplicates removed"""
        return {dependency.current for dependency in self.data}

    def __repr__(self) -> str:
        if len(self.versions) == 0:
            return f"No `{self.display_name}` versions found"
        if len(self.versions) == 1:
            return f"Only one `{self.display_name}` version found"
        suffix = ""
        if not self.has_all_expected_items:
            suffix = f" (only {len(self.versions)} of {self.expected_num} found)"
        if self.is_consistent:
            return f"Versions match for package `{self.display_name}`: {', '.join(self.versions)}{suffix}"
        return f"Versions mismatch for package `{self.display_name}`: {', '.join(self.versions)}{suffix}"



"""Model classes for lists of package metadata items."""
from .available_updates import AvailableUpdates
from .packages_metadata_list import PackagesMetadataList
from .versions_match import VersionsMatch
from .vulnerabilities import Vulnerabilities



"""Representation of a package manager. Use this class to obtain all packages managed by a particular package \
manager across all servers and directories."""
from dataclasses import dataclass
from enum import Enum
from typing import List, Union

from omniversion.models.package_metadata.list.packages_metadata_list import PackagesMetadataList
from omniversion.models.package_metadata.package_metadata import PackageMetadata


class PackageManagerIdentifier(Enum):
    """All package managers currently known to omniversion."""

    apt = "apt"
    """[Aptitude](https://salsa.debian.org/apt-team/apt)"""
    file = "file"
    """Not an actual package manager. We use this value for package infos extracted from local configuration files."""
    galaxy = "galaxy"
    """[Ansible Galaxy](https://galaxy.ansible.com)"""
    go = "go"
    """[go modules](https://go.dev/ref/mod)"""
    homebrew = "homebrew"
    """[homebrew](https://brew.sh)"""
    npm = "npm"
    """[npm](https://www.npmjs.com)"""
    nvm = "nvm"
    """[nvm](https://github.com/nvm-sh/nvm)"""
    pip = "pip"
    """[pip](https://pypi.org/project/pip/)"""
    rubygems = "rubygems"
    """[rubygems](https://rubygems.org)"""
    rvm = "rvm"
    """[rvm](https://rvm.io)"""


@dataclass
class PackageManager:
    """Representation of a package manager. Use this class to obtain all packages managed by a particular package \
    manager across all servers and directories."""

    identifier: str
    """The package manager's unique identifier. Refer to `PackageManagerIdentifier` for valid values."""
    packages: PackagesMetadataList

    @classmethod
    def list_from_packages(cls, packages: Union[PackagesMetadataList, List[PackageMetadata]]) -> List['PackageManager']:
        """Create a list of `PackageManager` objects from a list of package metadata items.
        
        Parameters
        ----------
        packages
            A `PackagesMetadataList` or a list of `PackageMetadata` items that should be be grouped by their \
            `package_manager` property.
            
        Returns
        -------
        List[PackageManager]
            A list of `PackageManager` objects, each containing only the package infos obtained via this package \
            manager.        
        """
        package_manager_identifiers = sorted(list({package.package_manager for package in packages
                                                   if package.package_manager is not None}))
        return [PackageManager(identifier, PackagesMetadataList([
            package for package in packages if package.package_manager == identifier]))
                for identifier in package_manager_identifiers]

    def __str__(self):
        return f"package manager: {self.identifier}"



from dataclasses import dataclass
from typing import Optional


@dataclass
class InstalledPackage:
    """InstalledPackage is an installation of a software package
        in a particular location on a particular host.
    """

    location: Optional[str] = None
    version: Optional[str] = None

    def __str__(self):
        """Debug string representation.

        Use the `omniversion.pretty_print` module for prettier output more suitable for human consumption."""
        version = '?' if self.version is None else self.version
        location = '?' if self.location is None else self.location
        return f"version `{version}` in location `{location}`"



"""A list of package metadata items, each describing all that is known about a particular software package."""
from collections import UserList
from typing import Optional, Union, List

from ..package_metadata import PackageMetadata

from .available_updates import AvailableUpdates
from .versions_match import VersionsMatch
from .vulnerabilities import Vulnerabilities


class PackagesMetadataList(UserList):
    """A list of package metadata items, each describing all that is known about a particular software package.

    This object is offers convenient methods for filtering packages."""
    def filter(
            self,
            verb: Optional[Union[str, List[str]]] = None,
            host: Optional[str] = None,
            package_manager: Optional[str] = None,
            package_name: Optional[Union[str, List[str]]] = None,
    ) -> 'PackagesMetadataList':
        """List all dependencies matching the given criteria.

        Parameters
        ----------
        verb
            If provided, will ensure only packages obtained via this verb/these verbs will be returned. A verb is a \
            package manager command type like `list`, `audit`, `outdated` etc.

        host
            If provided, will ensure only packages extracted from to the specified host(s) will be returned.

        package_manager
            If provided, the identifier of a package manager whose packages should be returned.

        package_name
            If provided, only packages matching this name or list of names will be returned.

        Returns
        -------
        PackagesMetadataList
            A list of matching packages.
        """
        def filter_condition(package: PackageMetadata) -> bool:
            if host is not None and package.host != host:
                return False
            if package_manager is not None and package.package_manager != package_manager:
                return False
            if isinstance(verb, list) and package.verb not in verb:
                return False
            if isinstance(verb, str) and verb != package.verb:
                return False
            if package_name is None:
                return True
            if isinstance(package_name, list):
                return package.name in package_name
            return package.name == package_name
        return PackagesMetadataList([item for item in self if filter_condition(item)])

    def audit(
            self,
            host: Optional[str] = None,
            package_manager: Optional[str] = None,
            package_name: Optional[Union[str, List[str]]] = None,
    ) -> Vulnerabilities:
        """Extract security vulnerabilities from the data.

        Parameters
        ----------
        host
            If provided, will ensure only packages extracted from to the specified host(s) will be returned.

        package_manager
            If provided, the identifier of a package manager whose packages should be returned.

        package_name
            If provided, only packages matching this name or list of names will be returned.

        Returns
        -------
        omniversion.models.package_metadata.list.vulnerabilities.Vulnerabilities
            All security issues contained in the input.
        """
        return Vulnerabilities(self.filter("audit", host, package_manager, package_name))

    def ls(
            self,
            host: Optional[str] = None,
            package_manager: Optional[str] = None,
            package_name: Optional[Union[str, List[str]]] = None,
    ):
        """Extract installed package metadata items.

        Parameters
        ----------
        host
            If provided, will ensure only packages extracted from to the specified host(s) will be returned.

        package_manager
            If provided, the identifier of a package manager whose packages should be returned.

        package_name
            If provided, only packages matching this name or list of names will be returned.

        Returns
        -------
        PackagesMetadataList
            All package metadata items relating to installed packages contained in the input.
        """
        return PackagesMetadataList(self.filter(["list", "version"], host, package_manager, package_name))

    def outdated(
            self,
            host: Optional[str] = None,
            package_manager: Optional[str] = None,
            package_name: Optional[Union[str, List[str]]] = None,
    ):
        """Extract available updates from the data.

        Parameters
        ----------
        host
            If provided, will ensure only packages extracted from to the specified host(s) will be returned.

        package_manager
            If provided, the identifier of a package manager whose packages should be returned.

        package_name
            If provided, only packages matching this name or list of names will be returned.

        Returns
        -------
        omniversion.models.package_metadata.list.available_updates.AvailableUpdates
            All available updates for the packages contained in the input.
        """
        return AvailableUpdates(self.filter("outdated", host, package_manager, package_name))

    def show(
            self, package_name: Union[str, List[str]], display_name: Optional[str] = None
    ):
        """Extract metadata for a particular package or list of related packages.

        This is especially useful for ensuring consistency of versions installed on different hosts, \
        as well as versions of related packages delivered through different package managers.

        Parameters
        ----------
        package_name
            The name(s) of the package(s) that should be extracted.

        display_name
            An optional value used in human-readable output.

        Returns
        -------
        omniversion.models.package_metadata.list.versions_match.VersionsMatch
            All available metadata for the specified package(s).
        """
        return VersionsMatch(
            self.filter(["list", "version"], package_name=package_name).data,
            package_name,
            display_name,
        )

    def __repr__(self) -> str:
        num_dependencies = len(self)
        return f"{num_dependencies} package{ '' if num_dependencies == 1 else 's'}"



"""A special `omniversion.models.host.host.Host` object denoting the machine on which the `omniversion` Python module \
is being run."""
from .host import Host
from ..package_metadata.list.packages_metadata_list import PackagesMetadataList


class Localhost(Host):
    """A special `omniversion.models.host.host.Host` object denoting the machine on which the `omniversion` Python \
    module is being run."""
    def __init__(self, packages: PackagesMetadataList):
        super().__init__("localhost", packages)

    def __str__(self):
        return "localhost"



"""A remote server, container or local host from which package data can be fetched."""
from dataclasses import dataclass
from typing import List, Union

from omniversion.models.package_metadata.package_metadata import PackageMetadata
from omniversion.models.package_metadata.list.packages_metadata_list import PackagesMetadataList


@dataclass
class Host:
    """A remote server, container or local host from which package data can be fetched."""
    name: str
    """The name of the host, e.g. an ssh host or `localhost`. This should be unique."""
    packages: PackagesMetadataList
    """The package data fetched from this host."""

    @classmethod
    def list_from_packages(cls, packages: Union[PackagesMetadataList, List[PackageMetadata]]) -> List['Host']:
        """Create a list of hosts from a list of packages.

        Parameters
        ----------
        packages
            A `PackagesMetadataList` or a list of `PackageMetadata` items that should be grouped by their `host` \
            property.

        Returns
        -------
        List[Host]
            A list of `Host` objects, each containing only the packages belonging to this host.
        """
        hostnames = sorted(list({package.host for package in packages}))
        return [Host(hostname, PackagesMetadataList([package for package in packages if package.host == hostname]))
                for hostname in hostnames]

    def __str__(self):
        return f"host: {self.name}"



"""An imported file including meta data"""
import os
import time
from dataclasses import dataclass
from typing import Optional, List, Any

import yaml
from dacite import from_dict, ForwardReferenceError, MissingValueError, UnexpectedDataError, WrongTypeError

from ..package_metadata.package_metadata import PackageMetadata

STALENESS_THRESHOLD_IN_SECS: int = 60 * 60

AVAILABLE_VERBS = ["audit", "list", "refresh", "outdated", "version"]


@dataclass
class FileDataSource:
    """An imported file including meta data"""
    num_packages: Optional[int] = None
    """The number of package metadata items that have been imported."""
    verb: Optional[str] = None
    """The type of package manager command used to create the data in this file (e.g. `list`, `audit`, etc.)."""
    host: Optional[str] = None
    """The host from which the data was extracted."""
    package_manager: Optional[str] = None
    """The identifier of the package manager through which the data was obtained."""
    name: Optional[str] = None
    """The file name without the path."""
    version: Optional[str] = None
    """The version of the `omniversion/cli` executable used to create the file."""
    timestamp: Optional[float] = None
    """The time the file was last changed. Used to determine if data is stale and should be re-fetched."""
    path: Optional[str] = None
    """The absolute or relative path to the file."""

    @property
    def is_stale(self):
        """Data that was fetched a long time ago is considered stale"""
        return self.timestamp is None or time.time() > STALENESS_THRESHOLD_IN_SECS + self.timestamp

    @property
    def has_data(self):
        """Does the file contain any parseable data at all?"""
        return self.num_packages is not None and self.num_packages > 0

    @classmethod
    def load_data(cls, file_path: str, verb: str, host: str, package_manager: str) -> \
            tuple[Optional['FileDataSource'], List[PackageMetadata]]:
        """Load data from the specified file.

        Parameters
        ----------
        file_path
            The relative or absolute path to the file.
        verb
            The type of package manager command used to create the data in this file (e.g. `list`, `audit`, etc.).
        host
            The host from which the data was extracted.
        package_manager
            The identifier of the package manager through which the data was obtained.

        Returns
        -------
        tuple[FileDataSource, List[omniversion.models.package_metadata.package_metadata.PackageMetadata]]
            A `FileDataSource` object containing extraction metadata and extracted package info as a list of \
            `omniversion.models.package_metadata.package_metadata.PackageMetadata` objects.
        """
        file_name = os.path.basename(file_path)
        if not os.path.exists(file_path):
            return None, []
        version, packages_data, timestamp = cls._extract_yaml_data(file_path)
        if packages_data is None:
            return FileDataSource(
                num_packages=None, version=version, name=file_name, host=host, package_manager=package_manager,
                verb=verb, timestamp=timestamp, path=file_path
            ), []
        else:
            packages: List[PackageMetadata] = [cls._parse_package_data(package_data) for package_data in packages_data]
            # remove packages that didn't parse correctly
            packages = [package for package in packages if package is not None]
            for package in packages:
                package.verb = verb
                package.host = host
                package.package_manager = package_manager
            return FileDataSource(
                num_packages=len(packages),
                version=version,
                name=file_name,
                host=host,
                package_manager=package_manager,
                verb=verb,
                timestamp=timestamp,
                path=file_path,
            ), packages

    @classmethod
    def _parse_package_data(cls, data: Any) -> Optional[PackageMetadata]:
        try:
            return from_dict(data_class=PackageMetadata, data=data)
        except (ForwardReferenceError, KeyError, MissingValueError, TypeError, UnexpectedDataError, WrongTypeError):
            # TODO: log this (!)
            return None

    @classmethod
    def _extract_yaml_data(cls, file_path: str) -> tuple[Optional[str], Optional[list[any]], Optional[float]]:
        """load an omniversion file containing yaml data"""
        timestamp = None
        try:
            timestamp = os.stat(file_path).st_ctime
            with open(file_path, encoding="utf8") as file:
                # a yaml file may contain multiple documents
                # we want to extract all
                documents = yaml.safe_load_all(file)
                # the version of omniversion
                version = None
                items = []
                for document in documents:
                    if document is None:
                        continue
                    # we assume that all documents within the same file were written by the same omniversion version
                    if version is None and "version" in document:
                        version = document["version"]
                    if "items" in document:
                        items += document["items"]
                return version, items, timestamp
        except (TypeError, yaml.YAMLError):
            # TODO: log this (!)
            return None, None, timestamp

    def __str__(self):
        if self.num_packages is None:
            return f"No valid `{self.verb}` data loaded for `{self.package_manager}`"
        data_recency = "Stale" if self.is_stale else "Recent"
        if self.num_packages == 1:
            return f"{data_recency} `{self.verb}` data loaded for `{self.package_manager}` (1 package)`"
        if self.num_packages > 1:
            return f"{data_recency} `{self.verb}` data loaded for `{self.package_manager}`" \
                   f" ({self.num_packages} packages)"
        return f"No `{self.verb}` data loaded for `{self.package_manager}`"



"""Result of a versions match, i.e. a list of installations with corresponding versions"""
from collections import UserList


class AvailableUpdates(UserList):
    """List of packages for which a newer version is available"""
    def __init__(self, initlist=None):
        super().__init__(initlist)

    def __repr__(self):
        num_updates = len([item for item in self if item.current is not None])
        num_not_installed = len([item for item in self if item.current is None])
        components = []
        if num_updates == 1:
            components.append(f"1 outdated package")
        if num_updates > 1:
            components.append(f"{num_updates} outdated packages")
        if num_not_installed == 1:
            components.append(f"1 package not installed")
        if num_not_installed > 1:
            components.append(f"{num_not_installed} packages not installed")
        if len(components) > 0:
            return ", ".join(components).capitalize()
        return "All packages up-to-date"



from dataclasses import dataclass
from typing import List, Optional


@dataclass
class PackageSource:
    identifier: Optional[str] = None
    url: Optional[str] = None
    versions: Optional[List[str]] = None



import os
import unittest

from omniversion import Omniversion
from omniversion.models import DataSources, FileDataSource, ConfigDataSource, PackageMetadata

TEST_FILE_PATH = os.path.join(os.path.dirname(__file__), "../vectors/test-env.txt")


class OmniversionTestCase(unittest.TestCase):
    def test_str_for_no_data(self):
        omniversion = Omniversion(base_path=None)
        self.assertEqual("omniversion data: no sources loaded", omniversion.__str__())

    def test_str_for_single_file(self):
        data_sources = DataSources()
        data_sources.files = [FileDataSource()]
        omniversion = Omniversion(base_path=None, data_sources=data_sources)
        self.assertEqual("omniversion data: 1 file and 0 config items loaded for 1 host", omniversion.__str__())

    def test_str_for_multiple_files(self):
        data_sources = DataSources()
        data_sources.files = [FileDataSource(), FileDataSource()]
        omniversion = Omniversion(base_path=None, data_sources=data_sources)
        self.assertEqual("omniversion data: 2 files and 0 config items loaded for 1 host", omniversion.__str__())

    def test_str_for_files_and_configs(self):
        data_sources = DataSources()
        data_sources.files = [FileDataSource(), FileDataSource()]
        data_sources.configs = [ConfigDataSource(file_path="/test", regex="")]
        omniversion = Omniversion(base_path=None, data_sources=data_sources)
        self.assertEqual("omniversion data: 2 files and 1 config item loaded for 1 host", omniversion.__str__())

    def test_hostnames(self):
        data_sources = DataSources()
        data_sources.files = [FileDataSource(name="1", host="test1"),
                              FileDataSource(name="2", host="test2"),
                              FileDataSource(name="2", host="test2")]
        data_sources.configs = [ConfigDataSource(host="test1", file_path="/test", regex=""),
                                ConfigDataSource(host="test2", file_path="/test", regex="")]
        omniversion = Omniversion(base_path=None, data_sources=data_sources)
        self.assertListEqual(["test1", "test2"], omniversion.hostnames)

    def test_localhost(self):
        data_sources = DataSources()
        data_sources.packages = [
            PackageMetadata(name="test1", host="host1"),
            PackageMetadata(name="test2", host="localhost"),
            PackageMetadata(name="test3", host="localhost"),
            PackageMetadata(name="test4", host="host1"),
        ]
        omniversion = Omniversion(base_path=None, data_sources=data_sources)
        localhost = omniversion.localhost
        self.assertEqual(2, len(localhost.packages))

    def test_load_test_vectors(self):
        omniversion = Omniversion(base_path=os.path.join(os.path.dirname(__file__), "./vectors"))
        self.assertGreater(len(omniversion.packages), 0)


if __name__ == '__main__':
    unittest.main()



import os
from typing import List, Optional

from .config import ConfigDataSource
from .file import FileDataSource
from ..package_metadata.package_metadata import PackageMetadata

VALID_VERBS = ["audit", "list", "refresh", "outdated", "version"]


class DataSources:
    """Metadata regarding the package info available to `omniversion`."""
    def __init__(self, files: Optional[List[FileDataSource]] = None, configs: Optional[List[ConfigDataSource]] = None,
                 packages: Optional[List[PackageMetadata]] = None) -> None:
        """Create a new data sources object, loading the specified files and configuration values.

        Parameters
        ----------
        files
            Optional list of `.omniverison.yaml` files from which package info will be loaded.
        configs
            Optional list of configuration files from which package info will be extracted.
        packages
            Additional packages to be included in addition to the loaded packages. These will have no \
        associated file.
        """
        self.packages = packages if packages is not None else []
        self.files = []
        if files is not None:
            for file in files:
                self.add_file(file_path=file.path, verb=file.verb, host=file.host, package_manager=file.package_manager)
        self.configs = []
        if configs is not None:
            for config in configs:
                self.add_config(file_path=config.file_path, regex=config.regex, name=config.name)

    def add_files(self, base_path: str) -> List[PackageMetadata]:
        """Load package info from a folder on the local machine.

        Parameters
        ----------
        base_path
            The base from which to load the files. The expected folder structure is \
        `base_path` / `hostname` / `package_manager` / `verb`.omniversion.yaml, \
        which is the structure emitted by the omniversion/ansible roles.

        Returns
        -------
        List
            A list of extracted package metadata is returned for convenience. The same data is also added \
        to the `packages` property of the `DataSources` object."""
        files = []
        packages = []
        real_path = os.path.realpath(base_path)
        host_dirs = [
            directory
            for directory in os.listdir(real_path)
            if os.path.isdir(os.path.join(real_path, directory))
        ]
        for host in host_dirs:
            host_path = os.path.join(real_path, host)
            package_manager_dirs = [
                directory
                for directory in os.listdir(host_path)
                if os.path.isdir(os.path.join(host_path, directory))
            ]
            for package_manager in package_manager_dirs:
                for verb in VALID_VERBS:
                    file_path = os.path.join(host_path, package_manager, verb + ".omniversion.yaml")
                    file, new_packages = FileDataSource.load_data(file_path, verb, host, package_manager)
                    if file:
                        files.append(file)
                    if new_packages:
                        packages += new_packages
        self.files.extend(files)
        self.packages.extend(packages)
        return packages

    def add_file(self, file_path: str, verb: str, host: str, package_manager: str) -> List[PackageMetadata]:
        """Add package info from a single `.omniversion.yaml` file.

        Parameters
        ----------
        file_path
            The relative or absolute path of the file to load. It is not required to have any particular structure.
        verb
            The verb, i.e. package manager command type, used to extract this package info (e.g. `list`/`audit`).
        host
            The name of the host from which the data was extracted.
        package_manager
            The identifier of the package manager from which the package info was obtained.

        Returns
        -------
        List[omniversion.models.package_metadata.package_metadata.PackageMetadata]
            A list of extracted package metadata for convenience. The same data is also added to the `packages` \
        property of the `DataSources` object.
        """
        file, packages = FileDataSource.load_data(file_path, verb, host, package_manager)
        if file:
            self.files.append(file)
        self.packages.extend(packages)
        return packages

    def add_config(self, file_path: str, regex: str, name: Optional[str] = None) -> List[PackageMetadata]:
        """Add package info from a configuration file in arbitrary format.

        Parameters
        ----------
        file_path
            The relative or absolute path of the file to load. It is not required to have any particular structure.
        regex
            A Python-flavor regular expression used to extract the version(s) from the file.
        name
            The name of the package. This is required if the regex does not contain a `name` group.

        Returns
        -------
        List[DataSources]
            A list of extracted package metadata for convenience. The same data is also added to the `packages` \
        property of the `DataSources` object.
        """
        config, packages = ConfigDataSource.load(file_path, regex, name)
        self.configs.append(config)
        self.packages.extend(packages)
        return packages

    @property
    def hostnames(self) -> List[str]:
        """Deduplicated list of hosts for which files are present in the list"""
        return sorted(list({file.host for file in self.files}))

    @property
    def host_infos(self) -> List[tuple[str, List[FileDataSource], List[ConfigDataSource]]]:
        """List of host information, containing the hostname, a list of `.omniversion.yaml` files and a list of \
        configuration files from which packages have been loaded.
        """
        def item_for_host(hostname: str):
            files = [file for file in self.files if file.host == hostname]
            configs = [config for config in self.configs if config.host == hostname]
            return hostname, files, configs
        return [item_for_host(hostname) for hostname in self.hostnames]

    @property
    def package_manager_identifiers(self) -> List[str]:
        """Deduplicated list of hosts for which files are present in the list."""
        return sorted(list({file.package_manager for file in self.files}))

    def __str__(self):
        num_files = len(self.files)
        num_configs = len(self.configs)
        num_hosts = len(self.hostnames)
        files_desc = f"{num_files} file{'s' if num_files != 1 else ''}"
        configs_desc = f"{num_configs} config item{'s' if num_configs != 1 else ''}"
        hosts_desc = f"{num_hosts} host{'s' if num_hosts != 1 else ''}"
        if num_files + num_configs == 0:
            return "no sources loaded"
        return f"{files_desc} and {configs_desc} loaded for {hosts_desc}"



"""Package info obtained from a configuration file.

Some package versions are specified in local files (e.g. a configuration file, an environment file, a Dockerfile,
or an Ansible playbook with a hardcoded version). You can extract them using regular expressions.
A ConfigDataSource tracks the metadata related to this process."""
import os
import re
from dataclasses import dataclass
from typing import Optional, List

from ...models.package_metadata.package_metadata import PackageMetadata


@dataclass
class ConfigDataSource:
    """Package info obtained from a configuration file.

    Some package versions are specified in local files (e.g. a configuration file, an environment file, a Dockerfile,
    or an Ansible playbook with a hardcoded version). You can extract them using regular expressions.
    A ConfigDataSource tracks the metadata related to this process."""
    file_path: str
    """The path of the configuration or environment file that specifies one or more package version(s)."""
    regex: str
    """A Python-flavor regular expression used to extract the version(s) from the file. It should contain a group 
    called `version` and, unless the `name` property is specified, a named group called `name`."""
    name: Optional[str] = None
    """The name of the package. This is required if the regex does not contain a `name` group."""
    num_packages: Optional[int] = None
    """After extraction, this value specifies the number of extracted versions."""
    host: Optional[str] = None
    """The host is given only for future compatibility. Currently, the only valid value is `localhost`."""

    @classmethod
    def load(cls, file_path: str, regex: str, name: Optional[str] = None) -> \
            tuple['ConfigDataSource', List[PackageMetadata]]:
        """Load package info from a local file

        Parameters
        ----------
        file_path
            A relative or absolute path to a file.
        regex
            A Python-flavor regular expression used to extract the version(s) from the file.
        name
            The name of the package. This is required if the regex does not contain a `name` group.

        Returns
        -------
        Tuple[ConfigDataSource, List[omniversion.models.package_metadata.package_metadata.PackageMetadata]]
            A `ConfigDataSource` object containing extraction metadata and extracted package info as a list of \
            `omniversion.models.package_metadata.package_metadata.PackageMetadata` objects.
        """
        real_file_path = os.path.realpath(file_path)
        with open(real_file_path, encoding="utf8") as file:
            try:
                matches = re.compile(regex).finditer(file.read())
                package_name = name
                packages = []
                for match in matches:
                    version = match.group("version")
                    if package_name is None:
                        package_name = match.group("name")
                    package_metadata = PackageMetadata(
                        host="localhost",
                        name=package_name,
                        package_manager="file",
                        current=version,
                    )
                    packages += [package_metadata]
                config_data_source = ConfigDataSource(
                    num_packages=len(packages),
                    host="localhost",
                    file_path=file_path,
                    regex=regex,
                    name=name,
                )
                return config_data_source, packages
            except IndexError:
                raise IndexError("Invalid regex. You need to provide a named group called `version` and either a name "
                                 "parameter or a named group called `name`")

    def __str__(self):
        def entries_text():
            if self.num_packages == 0:
                return "No packages"
            if self.num_packages == 1:
                return "1 package"
            return f"{self.num_packages} packages"
        return f"{entries_text()} found in config"



"""A list of package metadata items containing vulnerability notices"""
from collections import UserList


class Vulnerabilities(UserList):
    """A list of package metadata items containing vulnerability notices"""
    pass

    def __repr__(self):
        advisories = [[data_item, advisory] for data_item in self for advisory in data_item.advisories]
        if len(advisories) == 0:
            return "No vulnerabilities found"
        return f"{len(advisories)} vulnerabilities found" if len(advisories) > 1 else "One vulnerability found"



"""Root module. Starting point for loading and using omniversion data."""
from typing import List, Optional

from omniversion.models import PackagesMetadataList, Localhost, Host
from omniversion.models.data_source.data_sources import DataSources


class Omniversion:
    """The `omniversion.Omniversion` class used to load and handle all omniversion data.

    You will probably want to use an `omniversion/ansible` role first. It will fetch package metadata
    from remote hosts, saving it to local `.omniversion.yaml` files.

    Parameters
    ----------
    base_path
        The absolute or relative path to the folder containing the `.omniversion.yaml` files. If `None` is \
    specified, the default value of `/tmp/omniversion` (matching the `omniversion/ansible` roles' default output \
    directory) will be used. The expected folder structure is \
    `base_path` / `hostname` / `package_manager` / `verb`.omniversion.yaml, \
    which is the structure emitted by the `omniversion/ansible` roles.

    data_sources
        Optional data sources to include in addition to the data found in the specified `base_path`. \
    You won't normally need this.
    """

    def __init__(self, base_path: Optional[str] = "/tmp/omniversion", data_sources: Optional[DataSources] = None):
        self.sources = data_sources if data_sources is not None else DataSources()
        if base_path is not None:
            self.sources.add_files(base_path=base_path)

    @property
    def packages(self) -> PackagesMetadataList:
        """List of all package metadata items that have been loaded."""
        return PackagesMetadataList(self.sources.packages)

    @property
    def hostnames(self) -> List[str]:
        """Names of hosts from which data has been loaded."""
        return self.sources.hostnames

    @property
    def hosts(self) -> List[Host]:
        """List of hosts from which data has been loaded."""
        return Host.list_from_packages(self.packages)

    @property
    def localhost(self) -> Localhost:
        """Convenience property to access the localhost."""
        return Localhost(PackagesMetadataList(self.packages.filter(host="localhost")))

    def __str__(self):
        """Debug string representation describing the loaded data.

        Use the `omniversion.pretty_print` module for prettier output more suitable for human consumption."""
        return f"omniversion data: {self.sources}"



"""Omniversion model classes"""
from .host import Host, Localhost
from .data_source import ConfigDataSource, FileDataSource
from .data_source.data_sources import DataSources
from .package_metadata import Advisory, InstalledPackage, PackageMetadata
from .package_metadata.list import AvailableUpdates, PackagesMetadataList, VersionsMatch, Vulnerabilities



"""A vulnerability notice"""
from dataclasses import dataclass
from typing import Optional


@dataclass
class Advisory:  # pylint: disable=too-many-instance-attributes
    """A vulnerability notice"""
    access: Optional[str] = None
    cvss_score: Optional[float] = None
    identifier: Optional[str] = None
    overview: Optional[str] = None
    patched_versions: Optional[str] = None
    recommendation: Optional[str] = None
    references: Optional[str] = None
    severity: Optional[str] = None
    title: Optional[str] = None
    url: Optional[str] = None
    vulnerable_versions: Optional[str] = None

    def __str__(self):
        """Debug string representation.

        Use the `omniversion.pretty_print` module for prettier output more suitable for human consumption."""
        if self.severity:
            return f"known vulnerability (severity `{self.severity}`)"
        else:
            return "known vulnerability"




"""Contains all available information about a particular software package,
including existing installations, security advisories and all sorts of other metadata.

If multiple versions of the same package need to be tracked
for a single combination of package manager and host,
this should be done using a single PackageMetadata structure.
By contrast, otherwise identical packages made available by different package managers
are considered different, as their metadata and/or content *may* indeed differ.
We also distinguish between package metadata on different hosts,
as installations are different and available versions
will depend on the configured repositories, firewall settings etc.
"""

from dataclasses import dataclass, field
from typing import List, Optional

from .advisory import Advisory
from .dependency_type import DependencyType
from .installed_package import InstalledPackage
from .package_source import PackageSource


@dataclass
class PackageMetadata:  # pylint: disable=too-many-instance-attributes
    """PackageMetadata contains all available information about a particular software package,
including existing installations, security advisories and all sorts of other metadata.

If multiple versions of the same package need to be tracked
for a single combination of package manager and host,
this should be done using a single PackageMetadata structure.

By contrast, otherwise identical packages made available by different package managers
are considered different.
We also distinguish between package metadata on different hosts,
as installations are different and available versions
will depend on the configured repositories, firewall settings etc.
"""

    name: Optional[str] = None
    """The identifier used to install the package."""
    aliases: Optional[List[str]] = field(default_factory=lambda: [])
    """Alternative names by which the package may be known or have been known"""

    package_manager: Optional[str] = None
    """The package manager through which the package can be installed.
    
    If a package is available through multiple package managers, we should use separate
    PackageMetadata objects to track them.
    """
    host: Optional[str] = None
    verb: Optional[str] = None
    """The name of the machine on which this package has been installed, requested 
    or otherwise tracked by a package manager.
    
    This may be `localhost`, a hostname (with or without schema) or the name of a docker container.
    While the naming is flexible, care should be taken to keep these identifiers unique.
    If a package is installed on multiple hosts, we should use separate
    PackageMetadata objects to track them.
    """
    installations: List[InstalledPackage] = field(default_factory=lambda: [])
    """Metadata on versions of the package installed on the current machine."""

    type: Optional[DependencyType] = None
    """The kind of dependency (`prod`, `dev`, `peer`)"""
    optional: Optional[bool] = None
    """True if the dependency need not be installed, None if unknown"""
    direct: Optional[bool] = None
    """Specifies whether the dependency has been required directly or transitively"""

    current: Optional[str] = None
    """The currently installed version of the package as reported by the package manager.
    
    E.g. what is reported by `npm ls <package>`, `rvm info` `pip show <package>`.
    """
    default: Optional[str] = None
    """The version the package manager would select by default.
    
    E.g. the version designated as `default` by `rvm ls`.
    """
    latest: Optional[str] = None
    """The most recent version of the package known to be available."""
    wanted: Optional[str] = None
    """The version of the package that would be installed
    based on the relevant constraints defined for this package manager.
    
    E.g. this is the highest version matching the range defined in `package.json` for `npm`.
    """
    locked: Optional[str] = None
    """The version specified in the relevant lock file for this package manager.
    
    E.g. the version defined in `package-lock.json` or `npm-shrinkwrap.json`.
    """
    extraneous: Optional[bool] = None
    """True if the package is not required, but installed.
    
    Installed optional packages do not count as extraneous."""
    missing: Optional[bool] = None
    """Missing is true if the package is required, but not installed."""

    dependencies: Optional[List[str]] = None
    """Packages directly required by this package at runtime."""
    devDependencies: Optional[List[str]] = None
    """Packages directly required by this package during development."""
    peerDependencies: Optional[List[str]] = None
    """Packages directly required, but not managed by this package.
    
    E.g. a plugin might add functionality to another package
    which is assumed to be installed independently of the plugin.
    """

    architecture: Optional[str] = None
    """The architecture for which the package was compiled.
    
    E.g. the architecture field reported by `rvm ls` or `apt list`.
    """
    author: Optional[str] = None
    """The author or list of authors reported by the package manager."""
    description: Optional[str] = None
    """The package description reported by the package manager."""
    homepage: Optional[str] = None
    """The package's homepage reported by the package manager."""
    license: Optional[str] = None
    """A license identifier as reported by the package manager.
    
    This is not currently standardized across package managers.
    """
    sources: List[PackageSource] = field(default_factory=lambda: [])
    """Sources through which the package is available, as reported by the package manager.
    
    E.g. the `sources` field in `apt list`.
    """
    advisories: List[Advisory] = field(default_factory=lambda: [])
    """Security notices on known vulnerabilities.
    
    E.g. the information contained in the output of `npm audit`.
    """

    def __str__(self):
        """Debug string representation.

        Use the `omniversion.pretty_print` module for prettier output more suitable for human consumption."""
        if self.name is None:
            return "unknown package name"
        version = ""
        if self.current is not None:
            version = f"@{self.current} installed"
        elif self.wanted is not None:
            version = f"@{self.wanted} wanted"
        via = f" via `{self.package_manager}`" if self.package_manager else ""
        return f"package `{self.name}`{version}{via}"



import pathlib
import os
from setuptools import setup, find_packages

VERSION = os.environ['VERSION']

DIR = pathlib.Path(__file__).parent
README = (DIR / "README.md").read_text()

setup(
    name="omniversion",
    version=VERSION,
    description="Omniversion Python integration",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/omniversion/omniversion",
    author="Layer9 GmbH",
    author_email="hello@layer9.berlin",
    license="AGPL-v3.0-only",
    packages=find_packages(),
    python_requires=">=3.10.0",
    install_requires=[
        "colorama",
        "dacite",
        "PyYAML",
    ],
)



"""Pretty-print some data in a human-readable string"""
from typing import List
from pprint import PrettyPrinter

from colorama import Fore, Back, Style

from omniversion import Omniversion
from omniversion.models import DataSources, Host, FileDataSource, ConfigDataSource, PackagesMetadataList, \
    AvailableUpdates, Vulnerabilities, VersionsMatch, PackageMetadata
from omniversion.models.package_manager import PackageManager


def header(text):
    """Text styled like a header"""
    return f"\n{Back.LIGHTBLACK_EX} {text} {Style.RESET_ALL}"


def traffic_light(text, status="green"):
    """Green, amber or red text with leading symbol"""
    if status == "green":
        return f"{Fore.GREEN}[✔]{Style.DIM} {text} {Style.RESET_ALL}"
    if status == "amber":
        return f"{Fore.YELLOW}[-]{Style.DIM} {text} {Style.RESET_ALL}"
    if status == "red":
        return f"{Fore.RED}[✘]{Style.DIM} {text} {Style.RESET_ALL}"
    return "-"


def multiline_indent(multiline_string: str, levels: int = 2):
    return " " * levels + ("\n" + " " * levels).join(multiline_string.splitlines())


class OmniversionPrettyPrinter(PrettyPrinter):
    def format(self, obj: object, context, depth: int, level: int) -> tuple[str, bool, bool]:
        def recursive_format(value: object):
            return self.format(value, context=context, depth=depth, level=level - 1)[0]

        if isinstance(obj, Omniversion):
            def host_summary(host, dependencies):
                return f"{recursive_format(host)}\n{multiline_indent(recursive_format(dependencies))}\n\n"

            return "\n".join([
                header('Data sources'),
                multiline_indent(recursive_format(obj.sources)),
                header('Dependency count'),
                *[multiline_indent(host_summary(host, host.packages.ls())) for host in obj.hosts],
                header('Available updates'),
                *[multiline_indent(host_summary(host, host.packages.outdated())) for host in obj.hosts],
                header('Security audit'),
                *[multiline_indent(host_summary(host, host.packages.audit())) for host in obj.hosts],
            ]), False, False

        if isinstance(obj, DataSources):
            def host_summary(host_info: tuple[str, List[FileDataSource], List[ConfigDataSource]]):
                sources = [recursive_format(file) for file in host_info[1]] + \
                          [recursive_format(config) for config in host_info[2]]
                formatted_sources = "\n".join(sources)
                return f"{Back.BLUE} {host_info[0]} {Style.RESET_ALL}\n{multiline_indent(formatted_sources)}\n"

            return "\n".join([host_summary(host_info) for host_info in obj.host_infos]), False, False

        if isinstance(obj, Host):
            return "\n".join([
                f"{Back.BLUE} {obj.name} {Style.RESET_ALL}",
            ]), False, False

        if isinstance(obj, FileDataSource):
            color = "green"
            if obj.is_stale or obj.num_packages == 0:
                color = "amber"
            if obj.num_packages is None:
                color = "red"
            return traffic_light(obj, color), False, False

        if isinstance(obj, ConfigDataSource):
            color = "green"
            if obj.num_packages == 0:
                color = "amber"
            return traffic_light(obj, color), False, False

        if isinstance(obj, PackagesMetadataList):
            num_dependencies = len(obj)
            pretty_deps = "dependenc" + ("y" if num_dependencies == 1 else "ies")
            return f"{num_dependencies} {Style.DIM}{pretty_deps}{Style.RESET_ALL}", False, False

        if isinstance(obj, AvailableUpdates):
            def update_description(update: PackageMetadata):
                if update.current is None:
                    latest_version_text = f" version {Fore.WHITE}{update.latest} {Style.RESET_ALL}" \
                        if update.latest else ""
                    return f"package {Fore.BLACK}{Back.WHITE} `{update.name}` {Style.RESET_ALL}{latest_version_text}" \
                           f" can be installed"
                return f"update for {Fore.BLACK}{Back.WHITE} `{update.name}` {Style.RESET_ALL} available: " \
                       f"{Fore.WHITE}{update.current}{Style.RESET_ALL} -> {Fore.WHITE}{update.latest}{Style.RESET_ALL}"

            def package_manager_summary(pm: PackageManager):
                updates_for_pm = AvailableUpdates(pm.packages.data)
                summary = traffic_light(updates_for_pm.__repr__(), "green" if len(updates_for_pm) == 0 else "red")
                return "\n".join([
                    f"{Back.CYAN} {pm.identifier} {Style.RESET_ALL}",
                    multiline_indent(summary),
                    *[multiline_indent(multiline_indent(update_description(update))) for update in updates_for_pm]
                ])

            package_managers = PackageManager.list_from_packages(obj.data)
            if len(package_managers) == 0:
                return traffic_light('no package managers found', 'amber'), False, False
            return "\n".join([package_manager_summary(pm) for pm in package_managers]) + "\n", False, False

        if isinstance(obj, Vulnerabilities):
            advisories = [[data_item, advisory] for data_item in obj for advisory in data_item.advisories]
            audit_items = [
                f'\t{(data_item.name or "").ljust(20)}'
                + f'\t{(data_item.host or "").ljust(12)}'
                + f'\t{(data_item.package_manager or "").ljust(12)}'
                + f'\t{advisory.severity or ""}'
                for [data_item, advisory] in advisories
            ]
            return traffic_light(
                f"{obj}\n"
                + "\n".join(audit_items),
                "green" if len(advisories) == 0 else "red"), True, False

        if isinstance(obj, VersionsMatch):
            table_items = [
                f'\t{(item.current or "").ljust(20)}'
                + f'\t{(item.host or "").ljust(12)}'
                + f'\t{(item.package_manager or "").ljust(12)}'
                + f'\t{item.name or ""}'
                for item in obj.data
            ]
            color = "green"
            if len(obj.versions) <= 1 or not obj.is_consistent or not obj.has_all_expected_items:
                color = "red"
            return traffic_light(obj.__str__() + "\n".join(table_items), color), False, False

        return super().format(obj, context, depth, level)


def pprint(obj: object, indent=1, width=80, depth=None, compact=False, stream=None):
    OmniversionPrettyPrinter(indent=indent, width=width, depth=depth, compact=compact, stream=stream) \
        .pprint(obj)


def pformat(obj: object, indent=1, width=80, depth=None, compact=False) -> str:
    return OmniversionPrettyPrinter(indent=indent, width=width, depth=depth, compact=compact).pformat(obj)
